from collections import defaultdict
from typing import NamedTuple, Tuple, List, Mapping, Hashable

import attr
import aiger
import funcy as fn
from aiger.aig import Node, Inverter, ConstFalse, AndGate, Input
from bidict import bidict


class SymbolTable(defaultdict):
    def __init__(self, func, *args, **kwargs):
        super().__init__(None, *args, **kwargs)
        self.func = func

    def __missing__(self, key: Node):
        if isinstance(key, ConstFalse):
            self[key] = -self[Inverter(key)]
        else:
            self[key] = self.func(key)
        return self[key]


class CNF(NamedTuple):
    clauses: List[Tuple[int]]
    input2lit: Mapping[str, int]
    output2lit: Mapping[str, int]
    comments: Tuple[str]


def aig2cnf(circ, *, outputs=None, fresh=None, force_true=True):
    """Convert an AIGER circuit to CNF via the Tseitin transformation."""
    if fresh is None:
        max_var = 0

        def fresh(_):
            nonlocal max_var
            max_var += 1
            return max_var

    circ = aiger.to_aig(circ, allow_lazy=True)
    assert len(circ.latches) == 0

    # Define Boolean Algebra over clauses.
    clauses, gate2lit = [], SymbolTable(fresh)

    @attr.s(auto_attribs=True, frozen=True)
    class LitWrapper:
        lit: Hashable
        gate: Node

        @fn.memoize
        def __and__(self, other):
            gate = AndGate(self.gate, other.gate)
            wrapped = LitWrapper(gate2lit[gate], gate)

            out, left, right = wrapped.lit, self.lit, other.lit
            clauses.append((-left, -right, out))     # (left /\ right) -> out
            clauses.append((-out, left))             # out -> left
            clauses.append((-out, right))            # out -> right
            return wrapped

        def __invert__(self):
            gate = Inverter(self.gate)
            gate2lit[gate] = -self.lit
            return LitWrapper(gate2lit[gate], gate)

    def lift(obj) -> LitWrapper:
        assert isinstance(obj, (Input, bool))
        if isinstance(obj, bool):
            assert not obj
            obj = ConstFalse()

        return LitWrapper(gate2lit[obj], obj)

    # Interpret circ over Lit Boolean Algebra.
    inputs = {i: aiger.aig.Input(i) for i in circ.inputs}
    out2lit, _ = circ(inputs=inputs, lift=lift)
    out2lit = {k: v.lit for k, v in out2lit.items()}  # Remove Lit wrapper.
    in2lit = bidict({i: gate2lit[aiger.aig.Input(i)] for i in circ.inputs})

    # Force True/False variable to be true/false.
    if ConstFalse() in gate2lit:
        clauses.append((-gate2lit[ConstFalse()],))

    # Force outputs to appear as positive variables.
    for name, gate in circ.node_map.items():
        if not isinstance(gate, aiger.aig.Inverter):
            continue

        oldv = out2lit[name]
        newv = out2lit[name] = fresh(gate)
        clauses.append((-newv,  oldv))
        clauses.append((newv,  -oldv))

    if force_true:
        if outputs is None:
            outputs = circ.outputs

        for name in outputs:
            clauses.append((out2lit[name],))

    return CNF(clauses, in2lit, out2lit, circ.comments)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger_cnf']

package_data = \
{'': ['*']}

install_requires = \
['bidict>=0.21.0,<0.22.0', 'funcy>=1.12,<2.0', 'py-aiger>=6.0.0,<7.0.0']

setup_kwargs = {
    'name': 'py-aiger-cnf',
    'version': '5.0.4',
    'description': 'Python library to convert between AIGER and CNF',
    'long_description': "# py-aiger-cnf\nPython library to convert between AIGER and CNF\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-cnf/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-cnf)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger-cnf/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger-cnf)\n[![PyPI version](https://badge.fury.io/py/py-aiger-cnf.svg)](https://badge.fury.io/py/py-aiger-cnf)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n<!-- markdown-toc start - Don't edit this section. Run M-x markdown-toc-generate-toc again -->\n**Table of Contents**\n\n- [Installation](#installation)\n- [Usage](#usage)\n\n<!-- markdown-toc end -->\n\n# Installation\n\nIf you just need to use `aiger_cnf`, you can just run:\n\n`$ pip install py-aiger-cnf`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n# Usage\n\nThe primary entry point for using `aiger_cnf` is the `aig2cnf`\nfunction which, unsurprisingly, maps `AIG` objects to `CNF` objects.\n\n```python\nimport aiger\nfrom aiger_cnf import aig2cnf\n\nx, y, z = map(aiger.atom, ('x', 'y', 'z'))\nexpr = (x & y) | ~z\ncnf = aig2cnf(expr.aig)\n```\n\nNote that this library also supports `aiger` wrapper libraries so long\nas they export a `.aig` attribute. Thus, could also\nwrite:\n\n```python\ncnf = aig2cnf(expr)\n```\n\n\nThe `CNF` object is a `NamedTuple` with the following three fields:\n\n1. `clauses`: A list of tuples of ints, e.g., `[(1,2,3), (-1,\n   2)]`. Each integer represents a variable's id, with the sign\n   indicating the polarity of the variable.\n2. `input2lit`: A [bidict](https://bidict.readthedocs.io/en/master/)\n   from input names to variable ids.\n2. `output2lit`: A [bidict](https://bidict.readthedocs.io/en/master/)\n   from output names to variable ids.\n",
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/py-aiger-cnf',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import aiger
import funcy as fn
import hypothesis.strategies as st
from aiger import hypothesis as aigh
from hypothesis import given
from pysat.solvers import Glucose3

from aiger_cnf import aig2cnf


@given(aigh.Circuits, st.data())
def test_aig2cnf(circ, data):
    expr1 = aiger.BoolExpr(circ.unroll(1))
    cnf = aig2cnf(expr1)
    g = Glucose3()
    for c in cnf.clauses:
        g.add_clause(c)

    test_input = {i: data.draw(st.booleans()) for i in expr1.inputs}
    assumptions = []
    for name, val in test_input.items():
        if name not in cnf.input2lit:
            continue
        sym = cnf.input2lit[name]
        if not val:
            sym *= -1
        assumptions.append(sym)

    assert expr1(test_input) == g.solve(assumptions=assumptions)


def test_fresh():
    expr = aiger.atom('x') | aiger.atom('y')
    count = -1

    def fresh(_):
        nonlocal count
        count -= 1
        return count

    cnf = aig2cnf(expr, fresh=fresh)
    assert all(x < 1 for x in cnf.output2lit.values())


def test_force_true():
    expr = aiger.atom('x') | aiger.atom('y')

    cnf1 = aig2cnf(expr, force_true=True)
    cnf2 = aig2cnf(expr, force_true=False)
    assert set(cnf1.clauses) > set(cnf2.clauses)
    assert len(cnf1.clauses) == len(cnf2.clauses) + 1


def test_negation():
    x, y, z = aiger.atoms('x', 'y', 'z')
    expr1 = ~(x | y) & z
    expr2 = ~expr1

    count = 0

    @fn.memoize
    def fresh(x):
        nonlocal count
        count += 1
        return count

    cnf1 = aig2cnf(expr1, fresh=fresh, force_true=False)
    cnf2 = aig2cnf(expr2, fresh=fresh, force_true=False)
    assert set(cnf1.clauses) < set(cnf2.clauses)

    new_clauses = set(cnf2.clauses) - set(cnf1.clauses)
    new_clauses = {frozenset(x) for x in new_clauses}

    assert len(new_clauses) == 2

    # (5 => -6) /\ (-5 => 6)
    assert new_clauses == {frozenset([-5, -6]), frozenset([5, 6])}
